#!/usr/bin/env python
####################################################################################
# Copyright (c) 2009-2012, Centre for Advanced Internet Architectures
# Swinburne University of Technology, Melbourne, Australia
# (CRICOS number 00111D).
#
####################################################################################
#
# MRT slice: Tool for cutting out slices of a large MRT dump, 
# using MRT RIB dumps to create updates for the initial RIB 
# propagation of a BGP session.
# Requires an 4 byte AS number and TableDumpV2 capable dpkt version.
# See README-mrtslice for details.
#
####################################################################################
#
# This software was developed by Mattia Rossi <mrossi@swin.edu.au>
#
# This software has been made possible in part by a grant from
# the Cisco University Research Program Fund at
# Community Foundation Silicon Valley
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. The names of the authors, the "Centre for Advanced Internet Architecture"
#    and "Swinburne University of Technology" may not be used to endorse
#    or promote products derived from this software without specific
#    prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
###############################################################################

import dpkt
import struct
import socket
import re
import time
import glob
import gzip,bz2
from optparse import OptionParser

def main():
    try:
        import psyco
        psyco.full()
    except ImportError:
        pass
        

    parser = OptionParser()
    parser.add_option("-i", "--infile", dest="indump",default=0, action="store", 
                  help="MRT input dump file - add multiple files by specifying them as list separated by \",\" e.g.: -i file1,file2,file3. Filenames can include wildcards", metavar="DUMPFILE")
    parser.add_option("-o", "--outfile",  dest="outdump", default=0, action="store", 
                  help="MRT output dump file", metavar="DUMPFILE")
    parser.add_option("-r", "--ribdump",  dest="ribdump", default=0, action="store", 
                  help="MRT RIB dump file (TableDump_v2). The new MRT dump file will start from the time of the first RIB dump entry. If not set, the slice will start from the beginning of the input dump file, or from the time specified with the -s option", metavar="RIBDUMP")
    parser.add_option("-a", "--asipfile",  dest="asip", default="", action="store", 
                  help="ASIP file to determine whether the (TableDump_v2) entries in the RIB should be converted to 4-byte or 2-byte ASes to mathc the input files. Required if -r and -i are used")
    parser.add_option("-t", "--timespan",  dest="timespan", default="",type="string", action="store", 
                  help="Timespan of the slice in seconds. Add \"m\" for minutes, \"h\" for hours, \"d\" for days and \"w\" for weeks. If there is only a single input file and no RIB file or start time, the default is 1h. Otherwise the default is to join the RIB file and all input files and process them to the end.")
    parser.add_option("-s", "--starttime",  dest="starttimestring", default="", type="string", action="store", 
                  help="Start time of the slice if no RIB dump is given. The time needs to be in the format: \"dd/mm/YYYY HH:MM:SS\" for the GMT Timezone. If a RIB dump is given, it will override this setting.")
    (options, args) = parser.parse_args()
    
    dumpfilenum=0
    dumpfiles=[]
    out=0
    rib=0
    
    slicetime=0
    starttime=0
    endtime=0
    
    mrth=0
    mrt_h=0

    ribonly=0
    
    # Variables to find out the filetype
    bz2file=1
    gzipfile=1
    otherfile=1
    
    #AS-IP information store
    asip={}
    
    # Check all command line options for validity
    if(options.outdump==0):
        print "No output file given"
        parser.print_help()
        exit(1)
    else:
	try:
	  out=open(options.outdump, "w")
        except:
          parser.error("Could not open output file for writing")
          
    if (options.indump==0):
        if (options.ribdump==0):
            print "No input and no RIB file given!"
            parser.print_help()
            exit(1)
        else:
            ribonly=1
    else:
	if (options.ribdump==0):
	  if (options.starttimestring):
	    try:
	      starttime=time.mktime(time.strptime(options.starttimestring,"%d/%m/%Y %H:%M:%S"))
	    except ValueError:
	      parser.error("Not able to parse the time format passed with -s")
	else:
	  if (options.asip==""):
	    parser.error("ASIP file is needed, if Input file and RIB file are given")
	  else:
	    try:
	      asipf=open(options.asip,"r")
	    except:
	      parser.error("Could not open "+options.asip+" for reading")
	    else:
	      for line in asipf:
		line=line.strip()
		#we don't need the IP, just the AS number and whether it's 4-byte or 2-byte
		asip[int(line.split(" ")[1])]=int(line.split(" ")[2])
	dumpfilelist=options.indump.strip().split(",")
	if (dumpfilelist[-1]==""):
	    dumpfilelist=dumpfilelist[:-1]
	for dfile in dumpfilelist:
	  dfs=sorted(glob.glob(dfile))
	  dumpfiles=dumpfiles+dfs
	dump=bz2.BZ2File(dumpfiles[0])
	try:
	  dump.read(1024)
	except IOError:
	  bz2file=0
	else:
	  dump.close()
	  dump=bz2.BZ2File(dumpfiles[0])
	if (not bz2file):
	  dump=gzip.open(dumpfiles[0])
	  try:
	    dump.read(1024)
	  except IOError:
	    gzipfile=0
	  else:
	    dump.close()
	    dump=gzip.open(dumpfiles[0])
	if (not gzipfile and not bz2file):
	  dump=open(dumpfiles[0], "r")
        if (not dump):
            parser.error("Could not open first input file for reading")
        try:
            mrth=dpkt.mrt.MRTHeader(dump.read(dpkt.mrt.MRTHeader.__hdr_len__))
        except dpkt.dpkt.NeedData:
            parser.error("Input file appears not to be a valid MRT dump")
        if (mrth.type==dpkt.mrt.START):
            mrth=dpkt.mrt.MRTHeader(dump.read(dpkt.mrt.MRTHeader.__hdr_len__))
        if (mrth.type!=dpkt.mrt.BGP4MP):
            parser.error("Input file is not a valid MRT Update dump (BGP4MP)")
        else:
	    if (starttime==0):
		starttime=mrth.ts
        dump.close()
   
    # Parse the timespan string
    if (len(dumpfiles)==1 and options.timespan.strip()=="" and options.ribdump==0 and options.starttimestring==""):
	options.timespan="1h"
    if (not options.timespan.strip()==""):
      m=re.search("([0-9]+)([mhdw]*)", options.timespan)
      if(m):
	  slicetime=int(m.group(1))
	  if (m.group(2) == "m" ):
		  slicetime=slicetime*60
	  if (m.group(2) == "h" ):
		  slicetime=slicetime*60*60
	  if (m.group(2) == "d" ):
		  slicetime=slicetime*60*60*24
	  if (m.group(2) == "w" ):
		  slicetime=slicetime*60*60*24*7
	  if (slicetime == 0):
		  parse.error("The timespan entered is invalid (0). If you want to process all files to the end do not specify the -t parameter at all.")
      else:
	  parser.error("Error in timespan input")
    else:
	slicetime=0
    if (options.ribdump!=0):
	rib=bz2.BZ2File(options.ribdump)
	try:
	  rib.read(1024)
	except IOError:
	  bz2file=0
	else:
	  rib.close()
	  rib=bz2.BZ2File(options.ribdump)
	if (not bz2file):
	  rib=gzip.open(options.ribdump)
	  try:
	    rib.read(1024)
	  except IOError:
	    gzipfile=0
	  else:
	    rib.close()
	    rib=gzip.open(options.ribdump)
	if (not gzipfile and not bz2file):
	  rib=open(options.ribdump, "r")
        if (not rib):
            parser.error("Error opening ribdump file for reading")
        else:
            try:
                mrth=dpkt.mrt.MRTHeader(rib.read(dpkt.mrt.MRTHeader.__hdr_len__))
            except dpkt.dpkt.NeedData:
                parser.error("Ribdump file appears not to be a valid MRT dump")
            if (mrth.type!=dpkt.mrt.TABLE_DUMP_V2):
                parser.error("Ribdump file is not a valid TABLE_DUMP_V2 RIB dump")
            else:
                starttime=mrth.ts
 
    
    #Print the start and endtime in human readable format
    if ribonly:
        print "Converting RIB dump: "+options.ribdump+" to update dump: "+options.outdump
    else:
	if (slicetime):
	  endtime=starttime+slicetime
	  print "Extracting "+ str(slicetime) +" seconds MRT data from \"" + options.indump + "\" ("+ str(len(dumpfiles)) +" files in total)."
	  print "Start time is: " + time.strftime("%d/%m/%Y %H:%M:%S", time.localtime(starttime)) + " localtime ( GMT: " + str(starttime) +" - "+ time.strftime("%d/%m/%Y %H:%M:%S", time.gmtime(starttime)) +")."
	  print "End time is: " + time.strftime("%d/%m/%Y %H:%M:%S", time.localtime(endtime))+ " ( GMT: " + str(endtime) +" - "+ time.strftime("%d/%m/%Y %H:%M:%S", time.gmtime(endtime)) +")." 
	else:
	  print "Extracting all MRT data from \"" + options.indump + "\" ("+ str(len(dumpfiles)) +" files in total)."
	  print "Start time is: " + time.strftime("%d/%m/%Y %H:%M:%S", time.localtime(starttime)) + " localtime ( GMT: " + str(starttime) +" - "+ time.strftime("%d/%m/%Y %H:%M:%S", time.gmtime(starttime)) +")."
    
    updates=0 # variable to count the amount of processed update messages from update dump
    
    # Check whether we use a MRT RIB dump to start 
    # or whether we start from the beginning of the original MRT file
    # We have a RIB dump, so let's start with that
    if(rib!=0):
	print "Converting RIB entries to updates...  "
        # Extract the MRT packet
        while (mrth.subtype!=dpkt.mrt.TABLE_DUMP_V2_PEER_INDEX_TABLE):
            rib.read(mrth.len)
            mrth=dpkt.mrt.MRTHeader(rib.read(dpkt.mrt.MRTHeader.__hdr_len__))
        # first entry must be the peer index table
        peerindex=dpkt.mrt.TableDump2_PeerIndex(rib.read(mrth.len))
        
        count=0 # variable to count the amount of processed prefixes
        as4byte={} # dictionary to keep track of whether to use 2 byte encoding or 4 byte encoding for a specific AS
        
        # Extract the prefixes from the RIB 
        # We create one BGP message for each prefix for simplicity. This can be optimised
        mrth=dpkt.mrt.MRTHeader(rib.read(dpkt.mrt.MRTHeader.__hdr_len__))
        while (mrth):
            count=count+1
            if (count%4==0):
	      print "\r|",
	    if (count%4==1):
	      print "\r/",
	    if (count%4==2):
	      print "\r-",
	    if (count%4==3):
	      print "\r\\",
            # Create IPv4 messages
            if (mrth.subtype==dpkt.mrt.TABLE_DUMP_V2_RIB_IPV4_UNICAST or mrth.subtype==dpkt.mrt.TABLE_DUMP_V2_RIB_IPV4_MULTICAST):
                try:
                    ribv4=dpkt.mrt.TableDump2_IPV4(rib.read(mrth.len))
                except dpkt.dpkt.NeedData:
                    break
                for i in xrange(ribv4.entry_count):
		    # build 4byte-AS database
		    if peerindex.peers[ribv4.data[i].peer_index].asn not in asip:
		      continue
		    if peerindex.peers[ribv4.data[i].peer_index].asn not in as4byte:
		      for attr in ribv4.data[i].attributes:
			if (attr.type==dpkt.bgp.AS_PATH):
			  for seg in attr.data.segments:
			    #find returns >0 if not found, so this is true for an AS4Path:
			    if (repr(seg).find("ASPath")):
			      as4byte[peerindex.peers[ribv4.data[i].peer_index].asn]=1
			    else:
			      as4byte[peerindex.peers[ribv4.data[i].peer_index].asn]=0
                    # Create MRT, BGP4MP and BGP packet
                    # We adjust the RIB entries to what's in the update dump
                    attrlist=[]
                    if asip[int(peerindex.peers[ribv4.data[i].peer_index].asn)]==1:
		      #4 byte AS in the update dump
		      mrt=dpkt.mrt.MRTHeader()
		      mrt.type=dpkt.mrt.BGP4MP
		      mrt.subtype=dpkt.mrt.BGP4MP_MESSAGE_32BIT_AS
		      mrtsub=dpkt.mrt.BGP4MPMessage_32()
		      # Convert eventual 2 byte ASPath attribute in the RIB to 4 byte
		      for attr in ribv4.data[i].attributes:
			if (attr.type==dpkt.bgp.AS_PATH):
			  if (as4byte[peerindex.peers[ribv4.data[i].peer_index].asn]==0):
			    # 4-byte in update, 2-byte in RIB
			    # This case is a bit weird.
			    # 4-byte ASes would be encoded as 2-byte AS 23456 (AS_TRANS) in the rib
			    # and then 23456 would be encoded in 4-bytes here unless the 2-byte data
			    # collector also wrote the AS4_PATH attribute to the file... 
			    # So we'd need to check for the presence of that attribute first to 
			    # be able to reconstruct things properly. (Same for AGGREGATOR)
			    # Currently this is not needed, just leaving it in as a remainder.
			    newattr=u.Attribute(type=dpkt.bgp.AS_PATH,flags=64)
			    for oldseg in attr.data.segments:
			      newsegment=newattr.AS4Path.AS4PathSegment()
			      newsegment.type=oldseg.type
			      newsegment.len=oldseg.len  
			      newsegment.data=newsegment.path=oldseg.data
			      newattr.data=newattr.AS4Path(newsegment.pack())
			    attr=newattr
			    attr.len=len(attr.data)
			if (attr.type==dpkt.bgp.AGGREGATOR):
			  if (as4byte[peerindex.peers[ribv4.data[i].peer_index].asn]==0):
			    # 4-byte in update, 2-byte in RIB
			    newattr=u.Attribute(type=dpkt.bgp.AGGREGATOR,flags=192)
			    newattr.data=newattr.AS4Aggregator(ip=attr.data.ip,asn=attr.data.asn)
			    attr=newattr
			    attr.len=len(attr.data)
			attrlist.append(attr)
		    else:
		      #4 byte AS in the update dump
		      aspath4=0 #flag for adding the AS4_PATH attribute if needed
		      aggr4=0 #flag for adding the AS4_AGGREGATOR attribute if needed
		      mrt=dpkt.mrt.MRTHeader()
		      mrt.type=dpkt.mrt.BGP4MP
		      mrt.subtype=dpkt.mrt.BGP4MP_MESSAGE
		      mrtsub=dpkt.mrt.BGP4MPMessage()
		      # Convert eventual 2 byte ASPath attribute in the RIB to 4 byte
		      for attr in ribv4.data[i].attributes:
			if (attr.type==dpkt.bgp.AS_PATH):
			  if (as4byte[peerindex.peers[ribv4.data[i].peer_index].asn]==1):
			    # 2-byte in update, 4-byte in RIB
			    newattr=u.Attribute(type=dpkt.bgp.AS_PATH,flags=64)
			    segstring=''
			    for oldseg in attr.data.segments:
			      newsegment=newattr.ASPath.ASPathSegment(data=[])
			      newsegment.type=oldseg.type
			      newsegment.len=oldseg.len
			      for el in oldseg.data:
				if el <= 65535:
				  newsegment.data.append(el)
				else:
				  newsegment.data.append(23456)
				  aspath4=attr.data
				newsegment.path=newsegment.data
			      segstring+=newsegment.pack()
			    newattr.data=newattr.ASPath(segstring)
			    attr=newattr
			    attr.len=len(attr.data)
			if (attr.type==dpkt.bgp.AGGREGATOR):
			  if (as4byte[peerindex.peers[ribv4.data[i].peer_index].asn]==1):
			    # 4-byte in update, 2-byte in RIB
			    newattr=u.Attribute(type=dpkt.bgp.AGGREGATOR,flags=192)
			    if attr.data.asn <= 65535:
			      newattr.data=newattr.Aggregator(ip=attr.data.ip,asn=attr.data.asn)
			    else:
			      newattr.data=newattr.Aggregator(ip=attr.data.ip,asn=23456)
			      aggr4=attr.data
			    attr=newattr
			    attr.len=len(attr.data)
			attrlist.append(attr)
		      if aspath4:
			attr=u.Attribute(type=dpkt.bgp.AS4_PATH,flags=192)
			attr.data=aspath4
			attr.len=len(attr.data)
			attrlist.append(attr)
		      if aggr4:
			attr=u.Attribute(type=dpkt.bgp.AS4_AGGREGATOR,flags=192)
			attr.data=aggr4
			attr.len=len(attr.data)
			attrlist.append(attr)
		    mrt.ts=mrth.ts
		    mrtsub.dst_ip=peerindex.id
		    mrtsub.dst_as=0
		    mrtsub.intf=0
		    mrtsub.src_as=peerindex.peers[ribv4.data[i].peer_index].asn
		    mrtsub.src_ip=struct.unpack("!L",socket.inet_aton(peerindex.peers[ribv4.data[i].peer_index].address))[0]
		    bgppkt=dpkt.bgp.BGP(type=dpkt.bgp.UPDATE)
		    # Create Update container - need to put some things in it if we want to create it beforehand
		    # This creates an update message which withdraws the default route (0.0.0.0/0) four times
		    # The content is overwritten further down
		    u=bgppkt.Update("\x00\x04\x00\x00\x00\x00\x00\x00")
		    # Add attribute list
		    u.attributes=attrlist
                    # Add prefix from RIB entry
                    r=dpkt.bgp.RouteIPV4()
                    r.len = ribv4.prefix_len
                    r.prefix = ribv4.prefix
                    # Append prefix to announced list
                    u.announced.append(r)
                    # Delete the previously created withdrawn routes list
                    u.withdrawn=[]
                    # Add containers to their parent container,
                    # calculate length, pack mrt container and write to file
                    bgppkt.data=u
                    bgppkt.len=len(bgppkt)
                    mrtsub.data=bgppkt
                    mrt.data=mrtsub
                    mrt.len=len(mrtsub)
                    out.write(mrt.pack())
            # Create IPv6 messages
            if (mrth.subtype==dpkt.mrt.TABLE_DUMP_V2_RIB_IPV6_UNICAST or mrth.subtype==dpkt.mrt.TABLE_DUMP_V2_RIB_IPV6_MULTICAST):
                ribv6=dpkt.mrt.TableDump2_IPV6(rib.read(mrth.len))
		# not implemented at this stage
            
            # Insert UPDATE messages from original file
            # after one second of MRT recorded time
            # if the RIB dump spans more than one second
            #
            # This should  be hardly needed
            old_ts=mrth.ts
            try:
                mrth=dpkt.mrt.MRTHeader(rib.read(dpkt.mrt.MRTHeader.__hdr_len__))
            except dpkt.dpkt.NeedData:
                break
            if (ribonly==0):
                if(mrth.ts>old_ts):
		    print "RIB spans multiple seconds - Adding 1s of MRT updates from input files  "
		    for d in dumpfiles:
			dump=bz2.BZ2File(d)
			try:
			    dump.read(1024)
			except IOError:
			    bz2file=0
			else:
			    dump.close()
			    dump=bz2.BZ2File(d)
			if (not bz2file):
			    dump=gzip.open(d)
			    try:
				dump.read(1024)
			    except IOError:
				gzipfile=0
			    else:
				dump.close()
				dump=gzip.open(d)
			if (not gzipfile and not bz2file):
			    dump=open(d, "r")
			if (not dump):
			    print "Could not open input file \"" + d + "\" for reading"
			    exit(1)
			if (not mrt_h):
			    temp = dump.read(dpkt.mrt.MRTHeader.__hdr_len__)
			    if (len(temp) < dpkt.mrt.MRTHeader.__hdr_len__ ):
				break
			    mrt_h=dpkt.mrt.MRTHeader(temp)
			while (mrt_h.type != dpkt.mrt.BGP4MP):
			    temp = dump.read(dpkt.mrt.MRTHeader.__hdr_len__)
			    if (len(temp) < dpkt.mrt.MRTHeader.__hdr_len__ ):
				break
			    mrt_h=dpkt.mrt.MRTHeader(temp)
			while (mrt_h.ts<old_ts):
			    dump.read(mrt_h.len)
			    s=dump.read(dpkt.mrt.MRTHeader.__hdr_len__)
			    if (len(s)<dpkt.mrt.MRTHeader.__hdr_len__):
				break
			    mrt_h=dpkt.mrt.MRTHeader(s)
			while (mrt_h.ts==old_ts):
			    if (updates%4==0):
			      print "\r|",
			    if (updates%4==1):
			      print "\r/",
			    if (updates%4==2):
			      print "\r-",
			    if (updates%4==3):
			      print "\r\\",
			    out.write(mrt_h.pack())
			    out.write(dump.read(mrt_h.len))
			    updates=updates+1
			    temp = dump.read(dpkt.mrt.MRTHeader.__hdr_len__)
			    if (len(temp) < dpkt.mrt.MRTHeader.__hdr_len__ ):
				break
			    mrt_h=dpkt.mrt.MRTHeader(temp)
			if (mrt_h.ts>old_ts):
			    print
			    print "Resuming conversion of RIB to updates..  "
			    break

        # Finished adding RIB entries,  now proceed with update messages
        rib.close()
        print str(count) + " RIB entries converted to update messages - adding update data.."
        if(ribonly==0):
	    count=0
	    for d in dumpfiles:
		count+=1
		dump=bz2.BZ2File(d)
		try:
		  dump.read(1024)
		except IOError:
		  bz2file=0
		else:
		  dump.close()
		  dump=bz2.BZ2File(d)
		if (not bz2file):
		  dump=gzip.open(d)
		  try:
		    dump.read(1024)
		  except IOError:
		    gzipfile=0
		  else:
		    dump.close()
		    dump=gzip.open(d)
		if (not gzipfile and not bz2file):
		    dump=open(d, "r")
		if (not dump):
		    print "Could not open input file \"" + d + "\" for reading"
		    exit(1)
		print
		print "Processing file \"" + d + "\" (" + str(count)+ "/" + str(len(dumpfiles)) + ")"
		# Read MRT header for the timestamp, don't touch the rest.
		if (not mrt_h):
		    mrt_h=dpkt.mrt.MRTHeader(dump.read(dpkt.mrt.MRTHeader.__hdr_len__))
		print "Searching starting point..  "
		skip=0
		while (mrt_h.ts<mrth.ts):
		    if (skip%4==0):
		      print "\r|",
		    if (skip%4==1):
		      print "\r/",
		    if (skip%4==2):
		      print "\r-",
		    if (skip%4==3):
		      print "\r\\",
		    while (mrt_h.type != dpkt.mrt.BGP4MP):
			temp = dump.read(dpkt.mrt.MRTHeader.__hdr_len__)
			if (len(temp) < dpkt.mrt.MRTHeader.__hdr_len__ ):
			    break
			mrt_h=dpkt.mrt.MRTHeader(temp)
		    dump.read(mrt_h.len)
		    skip+=1
		    try:
			mrt_h=dpkt.mrt.MRTHeader(dump.read(dpkt.mrt.MRTHeader.__hdr_len__))
		    except dpkt.dpkt.NeedData:
			break
		print 
		print "Adding updates..  "
		while (mrt_h.ts>=mrth.ts):
		    if (updates%4==0):
		      print "\r|",
		    if (updates%4==1):
		      print "\r/",
		    if (updates%4==2):
		      print "\r-",
		    if (updates%4==3):
		      print "\r\\",
		    if (slicetime):
		      if (mrt_h.ts>endtime):
			break
		    while (mrt_h.type != dpkt.mrt.BGP4MP):
			temp = dump.read(dpkt.mrt.MRTHeader.__hdr_len__)
			if (len(temp) < dpkt.mrt.MRTHeader.__hdr_len__ ):
			    break
			mrt_h=dpkt.mrt.MRTHeader(temp)
		    out.write(mrt_h.pack())
		    out.write(dump.read(mrt_h.len))
		    updates=updates+1
		    try:
			mrt_h=dpkt.mrt.MRTHeader(dump.read(dpkt.mrt.MRTHeader.__hdr_len__))
		    except dpkt.dpkt.NeedData:
			break
		dump.close()
		print
	print str(updates) + " update messages added. File: " + options.outdump + " is ready to use."
        out.flush()
        out.close()
    # If there is no RIB dump, start from the beginning of the original MRT file
    # Must contain a valid RIB propagation sequence to create a useable 
    # set of updates
    else:
	updates=0
	if (len(dumpfiles)==1):
	    dump=bz2.BZ2File(dumpfiles[0])
	    try:
	      dump.read(1024)
	    except IOError:
	      bz2file=0
	    else:
	      dump.close()
	      dump=bz2.BZ2File(dumpfiles[0])
	    if (not bz2file):
	      dump=gzip.open(dumpfiles[0])
	      try:
		dump.read(1024)
	      except IOError:
		gzipfile=0
	      else:
		dump.close()
		dump=gzip.open(dumpfiles[0])
	    if (not gzipfile and not bz2file):
	      dump=open(dumpfiles[0], "r")
	    mrth=dpkt.mrt.MRTHeader(dump.read(dpkt.mrt.MRTHeader.__hdr_len__))
	    print "Searching starting point..  "
	    skip=0
	    while (mrth.ts<starttime):
		if (skip%4==0):
		  print "\r|",
		if (skip%4==1):
		  print "\r/",
		if (skip%4==2):
		  print "\r-",
		if (skip%4==3):
		  print "\r\\",
		while (mrth.type != dpkt.mrt.BGP4MP):
		    temp = dump.read(dpkt.mrt.MRTHeader.__hdr_len__)
		    if len(temp) < dpkt.mrt.MRTHeader.__hdr_len__ :
			break
		    mrth=dpkt.mrt.MRTHeader(temp)
		dump.read(mrth.len)
		skip+=1
		try:
		    mrth=dpkt.mrt.MRTHeader(dump.read(dpkt.mrt.MRTHeader.__hdr_len__))
		except dpkt.dpkt.NeedData:
		    break
	    print 
	    print "Found starting point at: " + str(mrth.ts) + " - Adding updates..  "
	    while (mrth):
		if (updates%4==0):
		  print "\r|",
		if (updates%4==1):
		  print "\r/",
		if (updates%4==2):
		  print "\r-",
		if (updates%4==3):
		  print "\r\\",
		if (slicetime):
		  if (mrth.ts>endtime):
		    break
		out.write(mrth.pack())
		out.write(dump.read(mrth.len))
		updates=updates+1
		try:
		    mrt_h=dpkt.mrt.MRTHeader(dump.read(dpkt.mrt.MRTHeader.__hdr_len__))
		except dpkt.dpkt.NeedData:
		    break
	    dump.close()
	    out.flush()
	    out.close()
	    print
	    print str(updates) + " update messages added. File: " + options.outdump + " is ready to use."
	else:
	    foundstart=0
	    for d in dumpfiles:
		dump=bz2.BZ2File(d)
		try:
		  dump.read(1024)
		except IOError:
		  bz2file=0
		else:
		  dump.close()
		  dump=bz2.BZ2File(d)
		if (not bz2file):
		  dump=gzip.open(d)
		  try:
		    dump.read(1024)
		  except IOError:
		    gzipfile=0
		  else:
		    dump.close()
		    dump=gzip.open(d)
		if (not gzipfile and not bz2file):
		    dump=open(d, "r")
		if (not dump):
		    print "Could not open input file \"" + d + "\" for reading"
		    exit(1)
		mrth=dpkt.mrt.MRTHeader(dump.read(dpkt.mrt.MRTHeader.__hdr_len__))
		print "Searching starting point..  "
		skip=0
		if (not foundstart):
		  while (mrth.ts<starttime):
		      if (skip%4==0):
			print "\r|",
		      if (skip%4==1):
			print "\r/",
		      if (skip%4==2):
			print "\r-",
		      if (skip%4==3):
			print "\r\\",
		      while (mrth.type != dpkt.mrt.BGP4MP):
			  temp = dump.read(dpkt.mrt.MRTHeader.__hdr_len__)
			  if len(temp) < dpkt.mrt.MRTHeader.__hdr_len__ :
			      break
			  mrth=dpkt.mrt.MRTHeader(temp)
		      dump.read(mrth.len)
		      skip+=1
		      try:
			  mrth=dpkt.mrt.MRTHeader(dump.read(dpkt.mrt.MRTHeader.__hdr_len__))
		      except dpkt.dpkt.NeedData:
			  break
		  if (mrth.ts>=starttime):
		    foundstart=1
		    print 
		    print "Found starting point at: " + str(mrth.ts) + " - Adding updates..  "
		else:
		    while (mrth):
			if (updates%4==0):
			  print "\r|",
			if (updates%4==1):
			  print "\r/",
			if (updates%4==2):
			  print "\r-",
			if (updates%4==3):
			  print "\r\\",
			if (slicetime):
			  if (mrth.ts>endtime):
			    break
			out.write(mrth.pack())
			out.write(dump.read(mrth.len))
			updates=updates+1
			try:
			    mrt_h=dpkt.mrt.MRTHeader(dump.read(dpkt.mrt.MRTHeader.__hdr_len__))
			except dpkt.dpkt.NeedData:
			    break

		dump.close()
		print
	    out.flush()
	    out.close()
	    print str(updates) + " update messages added. File: " + options.outdump + " is ready to use."
    
if __name__ == "__main__":
    main()
